// (c) Manuel Freiberger
// 2008-2011
// manuel.freiberger@gmx.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

// ==========================================================================
// USB-Bootloader host controller
//
// This program drives the USB-Bootloader from the host side.
// ==========================================================================

// $Id: bootloader.cpp 23 2010-05-04 23:39:56Z manuel $

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>

#include <libusb-1.0/libusb.h>

#include "hex_file.hpp"

const unsigned BOOTLOADER_PROJECT_ID = 0x0001;
const unsigned BOOTLOADER_VERSION    = 0x0005;  // version 0.5

const int VENDOR_REQUEST_TYPE_OUT = 0x40;
const int VENDOR_REQUEST_TYPE_IN = 0xC0;

const int VR_GET_VERSION   = 0x00;
const int VR_GET_STATUS    = 0x01;
const int VR_READ_PGM_CFG  = 0x02;
const int VR_WRITE_PGM_CFG = 0x03;
const int VR_ERASE_PGM_CFG = 0x04;
const int VR_READ_EEPROM   = 0x05;
const int VR_WRITE_EEPROM  = 0x06;
const int VR_RESET         = 0xFA;

//! This class handles the USB communication with a USB bootloader device.
class USBDevice
{
  public:
    //! Constructor. Initializes internal state of the object but does not
    //! communicate over the USB, yet. Call \p init prior to any other
    //! USB communication.
    USBDevice();

    //! The destructor will release the USB interface of the device.
    virtual ~USBDevice();

    //! Initialize USB communication.
    bool init();

    //! Reset the device.
    bool reset();

    //! Close the USB device. No further USB communication is allowed after
    //! calling this method.
    void close();

    //! Get the \p index-th string descriptor from your device. If successful,
    //! the method will return \p true and the descriptor is returned in
    //! \p descriptor.
    bool getStringDescriptor(const unsigned index, std::string& descriptor);

    //! Get the version of the firmware installed on the device. The method
    //! sends a 'GET VERSION' request to the device and returns the \p version.
    //! If the communication was successful, the method returns \p true.
    //! Otherwise, it will return \p false.
    bool getVersion(unsigned& version);

    //! Read \p length bytes from the EEPROM starting at \p offset into
    //! \p data. If successful, the method returns \p true.
    bool readEEPROM(const unsigned offset, const unsigned length,
                    std::vector<unsigned char>& data);

    bool writeEEPROM(const unsigned offset,
                     std::vector<unsigned char>& data);

    bool readFlashMemory(const unsigned offset, const unsigned length,
                         std::vector<unsigned char>& data);

    bool writeFlashMemory(const unsigned offset,
                          std::vector<unsigned char>& data);

    bool eraseFlashMemory(const unsigned offset, const unsigned length);

  private:
    //! USB context.
    libusb_context* m_usb_context;
    //! USB handle to our device.
    libusb_device_handle* m_usb_device_handle;
    //! Flag to store if the interface was claimed.
    bool m_interface_claimed;
    //! Timeout for USB transfers.
    int m_timeout;
    //! Timeout for writing data.
    int m_write_timeout;

    //! Method to get the id and version of a certain \p device.
    bool getIdVersion(libusb_device_handle* device,
                      std::vector<unsigned char>& result);

    //! Get the \p status data structure.
    bool getStatus(std::vector<unsigned char>& status, int timeout);
};

USBDevice::USBDevice()
  : m_usb_context(0),
    m_usb_device_handle(0),
    m_interface_claimed(false),
    m_timeout(100),
    m_write_timeout(5000)
{
}

USBDevice::~USBDevice()
{
  close();
  // close the lib
  if (m_usb_context)
    libusb_exit(m_usb_context);
}

bool USBDevice::init()
{
  // initialize the USB library
  if (m_usb_context == 0)
  {
    if (libusb_init(&m_usb_context) != 0)
      return false;
    libusb_set_debug(m_usb_context, 3);
  }
  // get all devices
  libusb_device** list;
  ssize_t num_devices = libusb_get_device_list(m_usb_context, &list);
  if (num_devices < 0)
    return false;

  // now look for our device
  for (ssize_t counter = 0; counter < num_devices; ++counter)
  {
    // get a device descriptor for the current device
    libusb_device_descriptor descriptor;
    if (libusb_get_device_descriptor(list[counter], &descriptor) != 0)
      continue;

    // check vendor and product id
    if ((descriptor.idVendor != 0x1C40) || (descriptor.idProduct != 0x07EC))
      continue;

    // at this point we have a device that is from a project; we have to
    // compare the project id and the version now;
    // but first try to open the device...
    libusb_device_handle* test_device = 0;
    if (libusb_open(list[counter], &test_device) != 0 || !test_device)
      continue;
    // try to claim interface 0
    if (libusb_claim_interface(test_device, 0) != 0)
    {
      // it did not work -> close the device and go on with the next one
      libusb_close(test_device);
      continue;
    }
    // get the version of the device
    std::vector<unsigned char> result;
    if (!getIdVersion(test_device, result) || result.size() != 8)
    {
      // could not get the version -> release the interface and the device
      libusb_release_interface(test_device, 0);
      libusb_close(test_device);
      continue;
    }
    bool match = true;
    std::vector<unsigned char>::const_iterator iter = result.begin();
    // bytes 0-3 have to be zero
    for (unsigned counter = 0; counter < 4; ++counter)
      match = match && (*iter++ == 0);
    // compare project id (lsb first)
    match = match && (*iter++ == (BOOTLOADER_PROJECT_ID & 0xFF));
    match = match && (*iter++ == ((BOOTLOADER_PROJECT_ID >> 8) & 0xFF));
    // compare project version (minor first)
    match = match && (*iter++ == (BOOTLOADER_VERSION & 0xFF));
    match = match && (*iter++ == ((BOOTLOADER_VERSION >> 8) & 0xFF));
    // test if all comparisons worked
    if (!match)
    {
      // matching failed -> release the interface and the device
      libusb_release_interface(test_device, 0);
      libusb_close(test_device);
      continue;
    }
    // at this point we have a valid open device
    m_usb_device_handle = test_device;
    m_interface_claimed = true;
    libusb_free_device_list(list, 1);
    return true;
  }
  // no matching device
  libusb_free_device_list(list, 1);
  return false;
}

bool USBDevice::reset()
{
  // reset the device; this command will timeout because the device resets
  // ifself and cannot respond to USB commands any longer!
  libusb_set_debug(m_usb_context, 0);
  libusb_control_transfer(m_usb_device_handle,
                          VENDOR_REQUEST_TYPE_OUT, VR_RESET,
                          0, 0, 0, 0, m_timeout);
  // try a simple reset
  if (libusb_reset_device(m_usb_device_handle))
    return true;
  // if the simple reset did not work, close the device and re-open it
  close();
  libusb_set_debug(m_usb_context, 3);
  return init();
}

void USBDevice::close()
{
  // release the interface if claimed
  if (m_interface_claimed && m_usb_device_handle)
  {
    libusb_release_interface(m_usb_device_handle, 0);
    m_interface_claimed = false;
  }
  // close the device
  if (m_usb_device_handle)
  {
    libusb_close(m_usb_device_handle);
    m_usb_device_handle = 0;
  }
}

bool USBDevice::getStringDescriptor(unsigned index, std::string& descriptor)
{
  unsigned char str_desc[256];
  int bytes_read = libusb_get_string_descriptor_ascii(
                     m_usb_device_handle, index, str_desc, 255);
  if (bytes_read >= 0)
  {
    descriptor = std::string((char*)str_desc, bytes_read);
    return true;
  }
  else
  {
    descriptor = "";
    return false;
  }
}

bool USBDevice::getVersion(unsigned& version)
{
  std::vector<unsigned char> temp;
  if (!getIdVersion(m_usb_device_handle, temp) || temp.size() != 8)
    return false;
  version = (unsigned(temp[7]) << 8) + unsigned(temp[6]);
  return true;
}

bool USBDevice::readEEPROM(const unsigned offset, const unsigned length,
                           std::vector<unsigned char>& data)
{
  data.resize(length);
  int bytes_read = libusb_control_transfer(
                     m_usb_device_handle,
                     VENDOR_REQUEST_TYPE_IN, VR_READ_EEPROM,
                     0, offset, &data[0], length, m_timeout);
  return (bytes_read == length);
}

bool USBDevice::writeEEPROM(const unsigned offset,
                            std::vector<unsigned char>& data)
{
  int bytes_sent = libusb_control_transfer(
                     m_usb_device_handle,
                     VENDOR_REQUEST_TYPE_OUT, VR_WRITE_EEPROM,
                     0, offset, &data[0], data.size(), m_timeout);
  if (bytes_sent != data.size())
    return false;

  // wait until there is no pending operation left
  std::vector<unsigned char> status;
  do
  {
    if (!getStatus(status, m_write_timeout))
      return false;
  } while (status[0] & 0x01);

  // check if the error bit is cleared
  return ((status[0] & 0x02) == 0);
}

bool USBDevice::readFlashMemory(const unsigned offset, const unsigned length,
                                std::vector<unsigned char>& data)
{
  data.resize(length);
  int bytes_read = libusb_control_transfer(
                     m_usb_device_handle,
                     VENDOR_REQUEST_TYPE_IN, VR_READ_PGM_CFG,
                     offset & 0xFFFF, (offset >> 16) & 0x00FF,
                     &data[0], length, m_timeout);
  return (bytes_read == length);
}

bool USBDevice::writeFlashMemory(const unsigned offset,
                                 std::vector<unsigned char>& data)
{
  int bytes_sent = libusb_control_transfer(
                     m_usb_device_handle,
                     VENDOR_REQUEST_TYPE_OUT, VR_WRITE_PGM_CFG,
                     offset & 0xFFFF, (offset >> 16) & 0x00FF,
                     &data[0], data.size(), m_timeout);
  if (bytes_sent != data.size())
    return false;

  // wait until there is no pending operation left
  std::vector<unsigned char> status;
  do
  {
    if (!getStatus(status, m_write_timeout))
      return false;
  } while (status[0] & 0x01);

  // check if the error bit is cleared
  return ((status[0] & 0x02) == 0);
}

bool USBDevice::eraseFlashMemory(const unsigned offset, const unsigned length)
{
  int bytes_sent = libusb_control_transfer(
                     m_usb_device_handle,
                     VENDOR_REQUEST_TYPE_OUT, VR_ERASE_PGM_CFG,
                     offset & 0xFFFF,
                     ((offset >> 16) & 0x00FF) | ((length & 0x00FF) << 8),
                     0, 0, m_timeout);
  if (bytes_sent != 0)
    return false;

  // wait until there is no pending operation left
  std::vector<unsigned char> status;
  do
  {
    if (!getStatus(status, m_write_timeout))
      return false;
  } while (status[0] & 0x01);

  // check if the error bit is cleared
  return ((status[0] & 0x02) == 0);
}

bool USBDevice::getIdVersion(libusb_device_handle* device,
                             std::vector<unsigned char>& result)
{
  result.resize(8);
  int bytes_read = libusb_control_transfer(
                     device,
                     VENDOR_REQUEST_TYPE_IN, VR_GET_VERSION,
                     0, 0, &result[0], 8, m_timeout);
  return (bytes_read == 8);
}

bool USBDevice::getStatus(std::vector<unsigned char>& status,
                          int timeout)
{
  status.resize(5);
  int bytes_read = libusb_control_transfer(
                     m_usb_device_handle,
                     VENDOR_REQUEST_TYPE_IN, VR_GET_STATUS,
                     0, 0, &status[0], 5, timeout);
  return (bytes_read == 5);
}

//! Print the \p data in hex format onto \p cout.
void printHex(const std::vector<unsigned char>& data)
{
  for (unsigned counter = 0; counter < data.size(); ++counter)
  {
    if (counter)
      std::cout << " ";
    std::cout << std::hex << std::setw(2) << std::setfill('0')
              << unsigned(data[counter]) << std::dec;
  }
}

bool convertRange(const std::string& range,
                  unsigned& start, unsigned& end);
bool compareFlash(USBDevice& usb_device, const std::string& filename);
bool eraseFlash(USBDevice& usb_device, const std::string& range);
bool readFlash(USBDevice& usb_device, const std::string& range,
               const std::string& filename);
bool writeFlash(USBDevice& usb_device, const std::string& filename);
void printHelp();

int main(int argc, char* argv[])
{
  // if there is only one argument and the argument is "-h" then print the
  // help message and quit
  if (argc == 2 && std::string(argv[1]) == "-h")
  {
    printHelp();
    return 0;
  }

  USBDevice usb_device;
  if (!usb_device.init())
  {
    std::cout << "Could not initialize a USB bootloader device!" << std::endl;
    std::cout << "Possible reasons:" << std::endl;
    std::cout << "  -The device is not connected" << std::endl;
    std::cout << "  -The device is not powered" << std::endl;
    std::cout << "  -You do not have permissions for accessing the USB"
              << std::endl;
    return -1;
  }

  std::cout << "USB bootloader device identification:" << std::endl;
  unsigned version;
  if (!usb_device.getVersion(version))
  {
    std::cout << "Error while reading bootloader version" << std::endl;
    return -1;
  }
  std::cout << "Bootloader version: " << ((version >> 8) & 0xFF)
            << '.' << (version & 0xFF) << std::endl;
  std::string manufacturer, product;
  if (!usb_device.getStringDescriptor(1, manufacturer)
      || !usb_device.getStringDescriptor(2, product))
  {
    std::cout << "Error while retrieving strings!" << std::endl;
    return -1;
  }
  std::cout << "Manufacturer:       " << manufacturer << std::endl;
  std::cout << "Product:            " << product << std::endl;
  std::cout << std::endl;

  std::cout << "Start parsing command-line" << std::endl;
  int counter = 1;
  while (counter < argc)
  {
    if (std::string(argv[counter]) == "-c")
    {
      ++counter;
      if (counter >= argc)
      {
        std::cout << "No file given after '-c' argument." << std::endl;
        return -1;
      }
      if (!compareFlash(usb_device, argv[counter]))
        return -1;
    }
    if (std::string(argv[counter]) == "-e")
    {
      ++counter;
      if (counter >= argc)
      {
        std::cout << "No range given after '-e' argument." << std::endl;
        return -1;
      }
      if (!eraseFlash(usb_device, argv[counter]))
        return -1;
    }
    if (std::string(argv[counter]) == "-r")
    {
      ++counter;
      if (counter >= argc)
      {
        std::cout << "No range given after '-r' argument." << std::endl;
        return -1;
      }
      std::string range = std::string(argv[counter]);
      ++counter;
      if (counter >= argc)
      {
        std::cout << "No file given after '-r' argument." << std::endl;
        return -1;
      }
      if (!readFlash(usb_device, range, argv[counter]))
        return -1;
    }
    if (std::string(argv[counter]) == "-w")
    {
      ++counter;
      if (counter >= argc)
      {
        std::cout << "No file given after '-w' argument." << std::endl;
        return -1;
      }
      if (!writeFlash(usb_device, argv[counter]))
        return -1;
    }
    if (std::string(argv[counter]) == "-q")
    {
      std::cout << "Resetting device" << std::endl;
      if (!usb_device.reset())
      {
        std::cout << "Error during device reset." << std::endl;
        return -1;
      }
    }
    if (std::string(argv[counter]) == "-h")
      printHelp();
    std::cout << std::endl;
    ++counter;
  }

  std::cout << "Done." << std::endl;

  return 0;
}

bool convertRange(const std::string& range, unsigned& start, unsigned& end)
{
  if (range.find(':') == std::string::npos
      || range.find(':') == 0
      || (range.find(':') + 1) == range.length())
  {
    std::cout << "Error wrong format for range. Use begin:end." << std::endl;
    return false;
  }
  {
    std::stringstream ss(range.substr(0, range.find(':')));
    ss >> std::hex >> start;
  }
  {
    std::stringstream ss(range.substr(range.find(':') + 1));
    ss >> std::hex >> end;
  }
  if (end < start)
  {
    std::cout << "Error: Range start must not be larger than range end."
              << std::endl;
    return false;
  }
  return true;
}

bool compareFlash(USBDevice& usb_device, const std::string& filename)
{
  std::cout << "Comparing flash against file '" << filename << "'."
            << std::endl;
  HexFileIn file(filename);
  if (!file.is_open())
  {
    std::cout << "Error opening '" << filename << "'." << std::endl;
    return false;
  }
  while (!file.eof())
  {
    unsigned address;
    std::vector<unsigned char> data;
    file.getLine(address, data);

    // read the memory
    std::vector<unsigned char> buffer(data.size());
    usb_device.readFlashMemory(address, data.size(), buffer);

    // compare
    for (unsigned counter = 0; counter < data.size(); ++counter)
    if (buffer[counter] != data[counter])
    {
      std::cout << "Error: Wrong value at flash address "
                << std::hex << std::setfill('0') << std::setw(4)
                << address << std::dec << "." << std::endl;
      return false;
    }
  }
  std::cout << "Comparison done." << std::endl;
  return true;
}

bool eraseFlash(USBDevice& usb_device, const std::string& range)
{
  unsigned start, end;
  if (!convertRange(range, start, end))
    return false;
  std::cout << "Erasing flash memory at "
            << std::hex << std::setfill('0') << std::setw(4) << start
            << ":" << std::setw(4) << end << std::dec << "." << std::endl;
  while (start <= end)
  {
    // calculate the amount of 64 byte blocks between end and start and
    // round up to the next higher integer
    unsigned length = (end - start + 64) / 64;
    if (length > 255)
      length = 255;
    usb_device.eraseFlashMemory(start, length);
    start += length * 64;
  }
  std::cout << "Flash erased." << std::endl;
  return true;
}

bool readFlash(USBDevice& usb_device, const std::string& range,
               const std::string& filename)
{
  unsigned start, end;
  if (!convertRange(range, start, end))
    return false;
  std::cout << "Reading flash at "
            << std::hex << std::setfill('0') << std::setw(4) << start << ":"
            << std::setw(4) << end
            << std::dec << " into file '" << filename << "'." << std::endl;
  HexFileOut file(filename);
  if (!file.is_open())
  {
    std::cout << "Error opening '" << filename << "'." << std::endl;
    return false;
  }
  file.setAddress(start);
  while (start <= end)
  {
    unsigned length = end - start + 1;

    if (length > 255)
      length = 255;
    std::vector<unsigned char> data;
    usb_device.readFlashMemory(start, length, data);
    file.addData(data);
    start += length;
  }
  std::cout << "Flash read." << std::endl;
  return true;
}

bool writeFlash(USBDevice& usb_device, const std::string& filename)
{
  std::cout << "Writing file '" << filename << "' to flash." << std::endl;
  HexFileIn file(filename);
  if (!file.is_open())
  {
    std::cout << "Error opening '" << filename << "'." << std::endl;
    return false;
  }
  while (!file.eof())
  {
    unsigned address;
    std::vector<unsigned char> data;
    file.getLine(address, data);
    usb_device.writeFlashMemory(address, data);

    // read back the memory
    std::vector<unsigned char> buffer(data.size());
    usb_device.readFlashMemory(address, data.size(), buffer);

    // compare
    for (unsigned counter = 0; counter < data.size(); ++counter)
    if (buffer[counter] != data[counter])
    {
      std::cout << "Error: Wrong value at flash address "
                << std::hex << std::setfill('0') << std::setw(4)
                << address << std::dec << "." << std::endl;
      std::cout << "Try to erase the flash memory first!" << std::endl;
      return false;
    }
  }
  std::cout << "File written." << std::endl;
  return true;
}

void printHelp()
{
  std::cout << "USB-Bootloader" << std::endl;
  std::cout << "Version: " << (BOOTLOADER_VERSION / 256) << "."
            << (BOOTLOADER_VERSION % 256) << std::endl;
  std::cout << std::endl;
  std::cout << "  -c file.hex             Compare flash memory against file"
            << std::endl;
  std::cout << "  -e start:end            Erase flash memory at [start:end]"
            << std::endl;
  std::cout << "  -r start:end file.hex   Read flash memory at [start:end]"
            << " into file" << std::endl;
  std::cout << "  -w file.hex             Write given file into flash"
            << std::endl;
  std::cout << "  -q                      Reset the device" << std::endl;
  std::cout << std::endl
            << "Numeric values have to be entered in hex format." << std::endl;
}

// End of $Id: bootloader.cpp 23 2010-05-04 23:39:56Z manuel $
