; =========================================================================
; Includes for the PIC18F2550 bootloader.
;
; (c) Manuel Freiberger
; 2008-2011
; manuel.freiberger@gmx.at
;
; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program.  If not, see <http://www.gnu.org/licenses/>.
;
; =========================================================================

; =========================================================================
; Variables.
; The bootloader needs some bytes to store its internal state and to pass
; arguments to functions. Further, a USB buffer for IO operations is
; required.
; =========================================================================

; USB control (6 bytes)
;   byte 0
;     bit 0    - flag set if in bootloader mode (RE3 = low at startup)
;     bit 1    - flag for 'remote wakeup' enabled
;     bit 7..2 - unused
;   byte 1 *
;     bit 1..0 - current status of the USB module
;                00 - USB module is powered up but has not received a reset cmd
;                01 - USB module was reset and is awaiting an address
;                10 - USB module got an address and waits for configuration
;                11 - USB module got a non-zero configuration cmd
;     bit 2    - unused
;     bit 3    - data is being received over USB and transfer is not complete,
;                yet
;     bit 4    - data is being sent over USB and transfer is not complete, yet
;     bit 5    - a descriptor is being sent over USB and transfer is still
;                active
;     bit 6    - address setup request has been received; after the status
;                stage the address will be set
;     bit 7    - error request flag
;   byte 2 *
;     content depends on active USB request:
;       a) address sent by the host
;       b) number of bytes to send during 'get descriptor' request
;   byte 3
;     offset in descriptor table while sending a descriptor
;   byte 4
;     the type of the descriptor to send together with its index
;     bits 7:6 - 00 ... device descriptor
;                01 ... configuration descriptor (index in bits 5:0)
;                10 ... string descriptor (index in bits 5:0)
;   byte 5
;     counter or temporary memory
;
; * When InitializeUSB is called, the bytes 1 and 2 of USB_CONTROL have a
; special meaning: They store the low (USB_CONTROL + 1) and the high
; (USB_CONTROL + 2) starting address of the USB buffer associated with
; endpoint 0. If you do not modify these bytes, the buffer will be placed
; at DEFAULT_USB_BUFFER_START.
USB_CONTROL EQU 00000000

; Pointer to USB vendor token handler (3 bytes)
;
; The bootloader offers the possibility to add vendor specific USB
; setup tokens for endpoint 0 and to handle USB activity on other endpoints.
; USB_HANDLER_FUNCTION is a three byte pointer to the function
; which handles these actions. The order of bytes is low, high, upper.
; In case endpoint 0 was addressed and a vendor-specific SETUP token
; shall be processed, FSR0 points to the start of the SETUP token's payload
; (that is to the beginning of the OUT buffer of endpoint 0). In all other
; cases FSR0 is not initialized. You can find out which endpoint was addressed
; by checking the value of USTAT. This handler function must not remove the
; activity from PIC's USB FIFO because the bootloader will do this itself.
; To enable this handler, also the bit USB_FIRMWARE_HANDLER_FLAG in
; USB_CONTROL has to be set.
; Input:
;   FSR0: Beginning of SETUP token's payload in case endpoint 0 was addressed.
; Output:
;   None. If the handler receives an unknown USB token, it has to set the
;   USB_ERROR_FLAG in byte 1 of USB_CONTROL.
USB_HANDLER_FUNCTION EQU 00000006
USB_HANDLER_FUNCTIONL EQU 00000006
USB_HANDLER_FUNCTIONH EQU 00000007
USB_HANDLER_FUNCTIONU EQU 00000008

; The bootloader module needs one more byte in data memory to store the
; packet size assigned to endpoint 0.
ENDPOINT0_PACKET_SIZE EQU 00000009

; The USB module requires two Buffer Descriptor Table entries for operation.
; BD0 is the BDT in outward direction, BD1 is in inward direction.
; Although these are not memory mapped registers, the PIC requires them to
; be at exactly those positions specified.
BD0STAT EQU 00000400
BD0CNT EQU 00000401
BD0ADRL EQU 00000402
BD0ADRH EQU 00000403
BD1STAT EQU 00000404
BD1CNT EQU 00000405
BD1ADRL EQU 00000406
BD1ADRH EQU 00000407

; This is the default starting address of the OUT-buffer for endpoint 0.
; The IN-buffer follows immediately. Both buffers have a size that is
; determined from the USB device descriptor. For the bootloader this is
; the value of DEFAULT_MAX_PACKET_SIZE0. You can relocate the buffer
; through setting bytes 1 and 2 of USB_CONTROL before calling the function
; InitializeUSB.
DEFAULT_USB_BUFFER_START EQU 00000408

; =========================================================================
; Constants
; =========================================================================

; The version of the Bootloader in BCD format (2 bytes).
BOOTLOADER_VERSION EQU 00000005

; USB_STATUS_MASK can be used to mask the status of the USB module from byte 1
; of USB_CONTROL
USB_STATUS_MASK EQU 00000003
; Constants for the status of the USB module
USB_STATUS_POWERED EQU 00000000
USB_STATUS_DEFAULT EQU 00000001
USB_STATUS_ADDRESSED EQU 00000002
USB_STATUS_CONFIGURED EQU 00000003
USB_RECV_DATA_FLAG EQU 00000003
USB_SEND_DATA_FLAG EQU 00000004
; Flag to set if an USB error is to be issued
USB_ERROR_FLAG EQU 00000007
USB_BOOTLOADER_FLAG EQU 00000000
; The USB_FIRMWARE_HANDLER_FLAG has to be set to install a firmware specific
; USB vendor request handler. If a USB vendor request is sent, the bootloader
; will jump to the address specified by USB_HANDLER_FUNCTION.
USB_FIRMWARE_HANDLER_FLAG EQU 00000002

; the size of the USB buffer in each direction
DEFAULT_MAX_PACKET_SIZE0 EQU 00000040

; To initialize a transfer over USB, BDnSTAT has to be set to either
; USB_TRANCEIVE_DATA0 or USB_TRANCEIVE_DATA1. Use the first one to
; transmit or receive data starting with a DATA0 packet. The latter one
; will do the same but starts with a DATA1 packet.
USB_TRANCEIVE_DATA0 EQU 00000088
USB_TRANCEIVE_DATA1 EQU 000000C8

; =========================================================================
; Functions
; =========================================================================

; ----------------------------------------------------------------------------
; Initialize the USB module.
;
; Input:
;   None
; Output:
;   None
; Modified:
;   w, FSR1
; Note:
;   This function will not yet set up the buffer descriptors completely.
;   It will only set the buffer addresses but not the BDnCNT or BDnSTAT
;   registers. This is done by HandleUSB as soon as a USB reset occurs. Such
;   a reset will be issued by the host during device initialization.
; ----------------------------------------------------------------------------
InitializeUSB EQU 000001B2

; -------------------------------------------------------------------------
; Call this function regularly, to handle USB activities.
; Input:
;   Vendor token handler in USB_HANDLER_FUNCTION
; Output:
;   None
; Modified:
;   w, FSR0, FSR1, BD0, BD1
;   This function will destroy TBLPTRU, TBLPTRH, TBLPTRL and TABLAT if
;   the device is in bootloader mode or if it is in firmware mode and no
;   firmware specific replacement for GetUSBDescriptor function is
;   supplied.
; -------------------------------------------------------------------------
HandleUSB EQU 000001F2

; -------------------------------------------------------------------------
; This is the bootloader's default function for filling the USB descriptor.
; If the device is not in bootloader mode, a 'goto 0x0804' is performed
; instead of calling this function.
; You can either provide your own function at address 0x0804 (which
; has to meet the specification of this function) or you can place a
; 'goto GetUSBDescriptor' there to use the bootloader's USB descriptors.
;
; Input:
;   USB_CONTROL + 3 ... the index of the byte in the descriptor to fetch.
;   USB_CONTROL + 4 ... the type of descriptor together with its index
;     bits 7:6 - 00 ... device descriptor
;                01 ... configuration descriptor (index in bits 5:0)
;                10 ... string descriptor (index in bits 5:0)
;   USB_CONTROL + 5 ... number of bytes to fetch (can be 0)
;   FSR1            ... pointer to the output buffer
; Output:
;   - The USB descriptor has been written to INDF1 (via FSR1).
;   - FSR1 and the register (USB_CONTROL + 3) have both been increased by the
;     number of bytes in register (USB_CONTROL + 5).
;   - Register (USB_CONTROL + 5) has been set to zero.
; Modified:
;   w, FSR1, TBLPTRx, TABLAT, USB_CONTROL + 3, USB_CONTROL + 5
; -------------------------------------------------------------------------
GetUSBDescriptor EQU 00000020

; ----------------------------------------------------------------------
; Send the next data packet over USB.
; Input:
;   FSR1            ... pointer to the data to send
;   USB_CONTROL + 2 ... the number of bytes to send
;   BD1STAT         ... contains the INVERTED DTS value
; Output:
;   None
; Note:
;   When calling this function, DTS (bit 6) in BD1STAT will be toggled.
;   That means that you have to clear DTS, if the first package is
;   to be transfered as DATA1 packet.
; ----------------------------------------------------------------------
SendUSBDataPacket EQU 0000072C

; ----------------------------------------------------------------------
; Receive another data packet over USB.
; Input:
;   FSR1            ... pointer to the receive buffer
;   USB_CONTROL + 2 ... the number of bytes to receive
;   BD0STAT         ... contain the INVERTED DTS value
; Output:
;   None
; Note:
;   When calling this function, DTS (bit 6) in BD0STAT will be toggled.
;   That means that you have to clear DTS, if the first package is
;   to be transfered as DATA1 packet.
; ----------------------------------------------------------------------
ReceiveUSBDataPacket EQU 00000760

; =========================================================================
; Start and interrupt vectors
; =========================================================================

; If RE.3 is high after the device starts, the bootloader is not active.
; In this case the program will continue at the location 0x0800. An offset
; of 0x0800 is also added to the interrupt vectors. This is their final
; location:
;   0x0800  start vector
;   0x0804  GetUSBDescriptor vector
;   0x0808  interrupt vector high
;   0x0818  interrupt vector low
