// (c) Manuel Freiberger
// 2008-2011
// manuel.freiberger@gmx.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

// ==========================================================================
// Interface for a Intel HEX file
// ==========================================================================

#ifndef HEXFILE_HPP
#define HEXFILE_HPP

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>

class HexFileIn
{
  public:
    HexFileIn(const std::string& filename)
      : m_filename(filename), m_file(), m_eof_record(false), m_line(0),
        m_address(0),
        m_linear_address_offset(0)
    {
      m_file.open(filename.c_str(), std::ifstream::in);
      readLine();
    }

    bool is_open() { return m_file.is_open(); }

    bool eof()
    {
      return m_file.eof() && m_eof_record;
    }

    void getLine(unsigned& address, std::vector<unsigned char>& data)
    {
      address = m_address;
      data = m_data;
      readLine();
    }

  private:
    std::string m_filename;
    std::ifstream m_file;
    // set to true if the EoF record was found
    bool m_eof_record;
    // currently processed line
    unsigned m_line;

    // current address
    unsigned m_address;
    unsigned m_linear_address_offset;
    // current data
    std::vector<unsigned char> m_data;

    bool readLine()
    {
      if (!m_file.is_open())
        return false;

      // process until a line with data was found
      bool data_line = false;

      do
      {
        if (m_file.eof() && m_eof_record)
          return true;

        if (m_file.eof() && !m_eof_record)
        {
          std::cout << "Error in line " << m_line
                    << ": Reached end of file before the EoF record."
                    << std::endl;
          return false;
        }

        std::string text;
        m_file >> text;
        ++m_line;
        if (!text.length())
          continue;

        // there must not be any data after the EoF record
        if (m_eof_record)
        {
          std::cout << "Error in line " << m_line
                    << ": Data past the EoF record!" << std::endl;
          return false;
        }
        // check start code
        if (text[0] != ':')
        {
          std::cout << "Error in line " << m_line << ": Wrong start code!"
                    << std::endl;
          return false;
        }
        // byte count
        unsigned byte_count;
        {
          std::stringstream ss(text.substr(1, 2));
          ss >> std::hex >> byte_count;
        }
        // address
        unsigned address;
        {
          std::stringstream ss(text.substr(3, 4));
          ss >> std::hex >> address;
        }
        // record type
        unsigned record_type;
        {
          std::stringstream ss(text.substr(7, 2));
          ss >> std::hex >> record_type;
        }
        // data
        m_data.resize(byte_count);
        for (unsigned counter  = 0; counter < byte_count; ++counter)
        {
          unsigned temp;
          std::stringstream ss(text.substr(9 + counter * 2, 2));
          ss >> std::hex >> temp;
          m_data[counter] = temp;
        }
        // checksum
        unsigned checksum;
        {
          std::stringstream ss(text.substr(9 + byte_count * 2, 2));
          ss >> std::hex >> checksum;
        }

        unsigned checksum_verify = byte_count + record_type
                                   + (address >> 8) + (address & 0xFF)
                                   + checksum;
        for (unsigned counter = 0; counter < byte_count; ++counter)
          checksum_verify +=  m_data[counter];

        if (checksum_verify & 0xFF)
        {
          std::cout << "Error in line " << m_line << ": Wrong checksum!"
                    << std::endl;
          return false;
        }

        // handle the record
        switch (record_type)
        {
          case 0: // data
            data_line = true;
            m_address = m_linear_address_offset + address;
          break;

          case 1: // EoF
            m_eof_record = true;
            break;

          case 4:
            if (address != 0)
            {
              std::cout << "Error in line " << m_line << ": Non-zero address."
                        << std::endl;
              return false;
            }
            if (byte_count != 2)
            {
              std::cout << "Error in line " << m_line
                        << ": Wrong number of bytes." << std::endl;
              return false;
            }
            m_linear_address_offset = (m_data[0] << 8) + m_data[1];
            m_linear_address_offset <<= 16;
            break;

          default:
            std::cout << "Error in line " << m_line << ": Unkown record type ("
                      << record_type << ")." << std::endl;
            return false;
        }

      } while (!data_line);
      return true;
    }
};

class HexFileOut
{
  public:
    HexFileOut(const std::string& filename)
      : m_filename(filename), m_file(), m_line(0), m_address(0),
        m_linear_address_offset(-1), m_buffer_pointer(0)
    {
      m_file.open(filename.c_str(), std::ofstream::out | std::ofstream::trunc);
    }

    ~HexFileOut()
    {
      flush();
      writeLine(0, 1, 0, 0);
    }

    bool is_open() { return m_file.is_open(); }

    void setAddress(unsigned address)
    {
      // if this is a new address, close the current line, first
      if (address != m_address + m_buffer_pointer)
        flush(true);
      m_address = address;
    }

    void addData(const std::vector<unsigned char>& data)
    {
      for (unsigned counter = 0; counter < data.size(); ++counter)
      {
        m_buffer[m_buffer_pointer++] = data[counter];
        flush();
      }
    }

  private:
    std::string m_filename;
    std::ofstream m_file;

    // the amount of lines written
    unsigned m_line;

    // address of the first byte in m_buffer
    unsigned m_address;
    // last linear address offset written to the file
    unsigned m_linear_address_offset;

    // a buffer plus a pointer in the buffer
    unsigned char m_buffer[16];
    unsigned m_buffer_pointer;

    void flush(bool force = false)
    {
      if (!m_buffer_pointer)
        return;

      // test if the address has still the same linear address offset
      if ((m_address >> 16) != m_linear_address_offset)
      {
        m_linear_address_offset = m_address >> 16;
        unsigned char offset[2];
        offset[0] = (m_linear_address_offset >> 8) & 0xFF;
        offset[1] = m_linear_address_offset & 0xFF;
        writeLine(0, 4, offset, 2);
      }

      // only write lines of 16 bytes except if we cross the linear address
      // boundary
      if (m_buffer_pointer == 16 || force
          || ((m_address + m_buffer_pointer) >> 16) != (m_address >> 16))
      {
        writeLine(m_address & 0xFFFF, 0, m_buffer, m_buffer_pointer);
        m_address += m_buffer_pointer;
        m_buffer_pointer = 0;
      }
    }

    void writeLine(unsigned address, unsigned record_type,
                   unsigned char* data, unsigned length)
    {
      unsigned checksum = length + ((address >> 8) & 0xFF)
                          + (address & 0xFF) + record_type;
      for (unsigned counter = 0; counter < length; ++counter)
        checksum += data[counter];
      checksum = ((checksum ^ 0xFF) + 1) & 0xFF;
      std::stringstream ss;
      ss << std::hex << std::setfill('0');
      ss << ":" << std::setw(2) << length
         << std::setw(2) << ((address >> 8) & 0xFF)
         << std::setw(2) << (address & 0xFF)
         << std::setw(2) << record_type;
      for (unsigned counter = 0; counter < length; ++counter)
        ss << std::setw(2) << (unsigned(data[counter]) & 0xFF);
      ss << std::setw(2) << checksum;
      m_file << ss.str() << std::endl;
      ++m_line;
    }
};

#endif
