// (c) Manuel Freiberger
// 2008-2011
// manuel.freiberger@gmx.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

// ==========================================================================
// Host program for USB interrupt transfer
//
// Used together with interrupt_transfer_firmware.asm
// ==========================================================================

// $Id: interrupt_transfer_control.cpp 20 2010-05-04 22:33:20Z manuel $

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <ctime>
#include <sys/time.h>

#include <libusb-1.0/libusb.h>

const int VENDOR_REQUEST_TYPE_OUT = 0x40;
const int VENDOR_REQUEST_TYPE_IN = 0xC0;

const int VR_GET_STATUS = 0x00;
const int VR_SET_LED = 0x01;
const int VR_CLEAR_LED = 0x02;
const int VR_RESET = 0x10;

void sleep_ms(unsigned millisec)
{
  clock_t end_time = millisec * CLOCKS_PER_SEC / 1000 + clock();
  while (clock() < end_time);
}

class USBDevice
{
  public:
    USBDevice();
    ~USBDevice();

    bool init();
    bool reset();

    bool getStringDescriptor(const unsigned index, std::string& descriptor);

    bool interruptTransfer();

  private:
    libusb_context* m_usb_context;
    libusb_device_handle* m_usb_device_handle;
    bool m_interface_claimed;

    int m_timeout;
};

USBDevice::USBDevice()
  : m_usb_context(0), m_usb_device_handle(0), m_interface_claimed(false),
    m_timeout(100)
{
}

USBDevice::~USBDevice()
{
  // release the interface if claimed
  if (m_interface_claimed)
    libusb_release_interface(m_usb_device_handle, 0);

  // close the device
  if (m_usb_device_handle)
    libusb_close(m_usb_device_handle);

  // close the lib
  if (m_usb_context)
    libusb_exit(m_usb_context);
}

bool USBDevice::init()
{
  // initialize the USB library
  if (libusb_init(&m_usb_context) != 0)
    return false;
  libusb_set_debug(m_usb_context, 3);

  // get all devices
  libusb_device** list;
  ssize_t num_devices = libusb_get_device_list(m_usb_context, &list);
  if (num_devices < 0)
    return false;

  // now look for our device
  libusb_device* device_found = 0;
  for (ssize_t counter = 0; counter < num_devices; ++counter)
  {
    // get a device descriptor for the current device_found
    libusb_device_descriptor descriptor;
    if (libusb_get_device_descriptor(list[counter], &descriptor) != 0)
    {
      libusb_free_device_list(list, 1);
      return false;
    }

    if ((descriptor.idVendor == 0x1C40) && (descriptor.idProduct == 0x07EC))
    {
      device_found = list[counter];
      break;
    }
  }
  if (!device_found)
  {
    libusb_free_device_list(list, 1);
    return false;
  }

  // open the device
  if (libusb_open(device_found, &m_usb_device_handle) != 0)
  {
    libusb_free_device_list(list, 1);
    return false;
  }
  // now we can free the list of devices (after opening!)
  libusb_free_device_list(list, 1);

  // claim interface 0
  if (libusb_claim_interface(m_usb_device_handle, 0) != 0)
    return false;
  else
    m_interface_claimed = true;

  return true;
}

bool USBDevice::reset()
{
  return libusb_reset_device(m_usb_device_handle) == 0;
}

bool USBDevice::getStringDescriptor(unsigned index, std::string& descriptor)
{
  unsigned char str_desc[256];
  int bytes_read = libusb_get_string_descriptor_ascii(
                     m_usb_device_handle, index, str_desc, 255);
  if (bytes_read >= 0)
  {
    descriptor = std::string((char*)str_desc, bytes_read);
    return true;
  }
  else
  {
    descriptor = "";
    return false;
  }
}

bool USBDevice::interruptTransfer()
{
  std::cout << "Testing transfer speed. Please be patient..." << std::endl;
  std::vector<unsigned char> data(64);
  for (unsigned counter = 0; counter < 64; ++counter)
    data[counter] = counter;

  timeval start_time, end_time;
  gettimeofday(&start_time, 0);
  for (unsigned iteration = 0; iteration < 1024; ++iteration)
  {
    int bytes_sent;
    int result = libusb_interrupt_transfer(
      m_usb_device_handle, 1, &data[0], data.size(), &bytes_sent, 1000);
    if (result != 0 || bytes_sent != data.size())
    {
      std::cerr << "Error during interrupt transfer." << std::endl;
      return false;
    }
  }
  gettimeofday(&end_time, 0);
  double diff_time = (double(end_time.tv_sec - start_time.tv_sec) * 1e6
                      + double(end_time.tv_usec - start_time.tv_usec)) / 1e6;
  std::cout << "Transfered 64kB in " << diff_time << " seconds." << std::endl;
  unsigned bytes_per_sec = unsigned(65536 / diff_time);
  std::cout << "This is a speed of " << bytes_per_sec << " bytes/second."
            << std::endl;
  double fraction = bytes_per_sec / 64e3;
  std::cout << "This is " << unsigned(fraction * 100) << "% of the maximum."
            << std::endl;

  return true;
}

int main(int argc, char* argv[])
{
  USBDevice usb_device;
  if (!usb_device.init())
  {
    std::cout << "Could not initialize the USB device!" << std::endl;
    std::cout << "Possibly you do not have the access rights for USB."
              << std::endl;
    return -1;
  }

  std::cout << "USB device identification:" << std::endl;
  std::string manufacturer, product;
  if (!usb_device.getStringDescriptor(1, manufacturer)
      || !usb_device.getStringDescriptor(2, product))
  {
    std::cout << "Error while retrieving strings!" << std::endl;
    return -1;
  }
  std::cout << "Manufacturer: " << manufacturer << std::endl;
  std::cout << "Product: " << product << std::endl;
  std::cout << std::endl;

  usb_device.interruptTransfer();

  return 0;
}

// End of $Id: interrupt_transfer_control.cpp 20 2010-05-04 22:33:20Z manuel $
