; $Id: interrupt_transfer_firmware.asm 20 2010-05-04 22:33:20Z manuel $

; ==========================================================================
;
; (c) Manuel Freiberger
; 2008-2011
; manuel.freiberger@gmx.at
;
; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program.  If not, see <http://www.gnu.org/licenses/>.
;
; ==========================================================================
; Example firmware for PIC18F2550 with USB bootloader
;
; This firmware implements USB interrupt transfer.
; ==========================================================================

; Processor specific stuff
    list p=18f2550
    include <p18f2550.inc>

include "bootloader.inc"

; Our own buffer descriptors for endpoint 1
BD2STAT EQU 0x408
BD2CNT  EQU 0x409
BD2ADRL EQU 0x40A
BD2ADRH EQU 0x40B

EP0_BUFFER_START EQU 0x420
EP1_BUFFER_START EQU 0x500
EP1_BUFFER_SIZE  EQU D'64'

YELLOW_LED_PIN EQU 0
RED_LED_PIN    EQU 1
GREEN_LED_PIN  EQU 6

; ---------------------------------------------------------------------------
; Interrupt and reset vectors.
; ---------------------------------------------------------------------------
    org 0x0800
    goto FirmwareMain           ; offset 0x00
    goto MyGetUSBDescriptor     ;        0x04
    goto $                      ;        0x08
    nop                         ;        0x0C
    nop                         ;        0x0E
    nop                         ;        0x10
    nop                         ;        0x12
    nop                         ;        0x14
    nop                         ;        0x16
    goto $                      ;        0x18

; ---------------------------------------------------------------------------
; Firmware-specific function for filling the buffer with USB descriptors
; ---------------------------------------------------------------------------
MyGetUSBDescriptor
    movlw upper(MyGetUSBDescriptor)
    movwf TBLPTRU, ACCESS
    movlw high(MyGetUSBDescriptor)
    movwf TBLPTRH, ACCESS
    ; extract the descriptor type
    movf USB_CONTROL + 4, W, ACCESS
    andlw 0xC0
    bnz mgud_check_configuration
    ; the device descriptor is wanted
    movlw low(DEVICE_DESCRIPTOR)
    bra mgud_add_index_and_fetch_byte
mgud_check_configuration
    xorlw 0x40
    bnz mgud_check_string
    ; a configuration descriptor is wanted; get the index
    movf USB_CONTROL + 4, W, ACCESS
    andlw 0x3F
    ; we only have one configuration descriptor; if the index is not zero
    ; then we issue an error
    bnz mgud_issue_USB_error
    movlw low(CONFIGURATION1_DESCRIPTOR)
    bra mgud_add_index_and_fetch_byte
mgud_check_string
    ; a string descriptor is wanted; get the index
    movf USB_CONTROL + 4, W, ACCESS
    andlw 0x3F
    bnz mgud_check_string_1
    ; string 0 is wanted
    movlw low(STRING0_DESCRIPTOR)
    bra mgud_add_index_and_fetch_byte
mgud_check_string_1
    xorlw 0x01
    bnz mgud_check_string_2
    ; string 1 is wanted
    movlw low(STRING1_DESCRIPTOR)
    bra mgud_add_index_and_fetch_byte
mgud_check_string_2
    xorlw (0x01 ^ 0x02)
    bnz mgud_issue_USB_error
    ; string 2 is wanted
    movlw low(STRING2_DESCRIPTOR)
    bra mgud_add_index_and_fetch_byte
mgud_issue_USB_error
    bsf USB_CONTROL + 1, USB_ERROR_FLAG, ACCESS
    return
mgud_add_index_and_fetch_byte
    ; add index
    addwf USB_CONTROL + 3, W, ACCESS
    movwf TBLPTRL, ACCESS
    ; increase the index
    movf USB_CONTROL + 5, W, ACCESS
    addwf USB_CONTROL + 3, F, ACCESS
    ; fetch the bytes in a loop
    incf USB_CONTROL + 5, F, ACCESS
mgud_loop
    dcfsnz USB_CONTROL + 5, F, ACCESS
    bra mgud_loop_done
    tblrd*+
    movf TABLAT, W, ACCESS
    movwf POSTINC1
    bra mgud_loop
mgud_loop_done
    return

; there has to be exactly one device descriptor for every device
; NOTE: a limitation of this bootloader is that only one configuration is
; supported which means that bNumConfigurations has to be 1
DEVICE_DESCRIPTOR
    db 0x12, 0x01           ; bLength, bDescriptorType
    db 0x00, 0x02           ; bcdUSB (2.0)
    db 0x00, 0x00           ; bDeviceClass (use interface class),bDeviceSubClass
       ; bDeviceProtocol (use interface protocol),
       ; bMaxPacketSize0 (low-speed: 8, full: 8/16/32/64)
    db 0x00, D'8'
    db 0x40, 0x1C           ; idVendor (0x1C40)
    db 0xEC, 0x07           ; idProduct (0x07EC)
    db 0x01, 0x00           ; bcdDevice (0.1)
    db 0x01, 0x02           ; iManufacturer, iProduct
    db 0x00, 0x01           ; iSerialnumber, bNumConfigurations
DEVICE_DESCRIPTOR_END

; there can be as many configuration descriptors per device as needed as long
; as the number of endpoints does not exceed 16; a configuration can have
; interfaces and endpoints (note that endpoint 0 does not need a descriptor
; as it is standardized)
; NOTE: a limitation of this bootloader is that the configuration descriptor
; must fit in 256 byte such that the high byte of wTotalLength is 0
CONFIGURATION1_DESCRIPTOR
    db 0x09, 0x02           ; bLength, bDescriptorType
    db D'25', 0x00          ; wTotalLength (low, high)
    db 0x01, 0x01           ; bNumInterfaces, bConfigurationValue
    db 0x00, 0x80           ; iConfiguration, bmAttributes (bus-powered)
    ; after bMaxPower the CONFIG1_INTERFACE1_DESCRIPTOR starts
    db D'50', 0x09          ; bMaxPower (in units of 2mA), bLength
    db 0x04, 0x00           ; bDescriptorType, bInterfaceNumber
    db 0x00, 0x01           ; bAlternateSetting, bNumEndpoints without EP0
    db 0xFF, 0x00           ; bInterfaceClass (vendor), bInterfaceSubClass
    db 0xFF, 0x00           ; bInterfaceProtokoll (vendor), iInterface
CONFIG1_INTERFACE1_DESCRIPTOR_END
; endpoint 0 needs no descriptor; the others do and look like
CONFIG1_ENDPOINT1_DESCRIPTOR
    db 0x07, 0x05           ; bLength, bDescriptorType
    db 0x01, 0x03           ; bEndpointAddress, bmAttributes
    db EP1_BUFFER_SIZE, 0x00    ; wMaxPacketSize (low, high)
    db 0x01                 ; bInterval
CONFIG1_ENDPOINT1_DESCRIPTOR_END
CONFIGURATION1_DESCRIPTOR_END

STRING0_DESCRIPTOR
    db 0x04, 0x03           ; bLength, bDescriptorType
    db 0x09, 0x04           ; wLANGID[0] (English - United States)
STRING0_DESCRIPTOR_END

STRING1_DESCRIPTOR
                            ; bLength, bDescriptorType
    db low(STRING1_DESCRIPTOR_END - STRING1_DESCRIPTOR), 0x03
    db 'C', 0x00
    db 'o', 0x00
    db 'm', 0x00
    db 'p', 0x00
    db 'a', 0x00
    db 'n', 0x00
    db 'y', 0x00
STRING1_DESCRIPTOR_END

STRING2_DESCRIPTOR
                            ; bLength, bDescriptorType
    db low(STRING2_DESCRIPTOR_END - STRING2_DESCRIPTOR), 0x03
    db 'I', 0x00
    db 'n', 0x00
    db 't', 0x00
    db 'e', 0x00
    db 'r', 0x00
    db 'r', 0x00
    db 'u', 0x00
    db 'p', 0x00
    db 't', 0x00
    db '-', 0x00
    db 'T', 0x00
    db 'e', 0x00
    db 's', 0x00
    db 't', 0x00
STRING2_DESCRIPTOR_END

    ; make sure that we did not cross a page boundary with the descriptor table
    if (high($) != high(MyGetUSBDescriptor))
        error "Descriptor definition crosses page boundary!"
    endif

; ----------------------------------------------------------------------
; This is the beginning of the firmware.
; ----------------------------------------------------------------------
FirmwareMain
    ; set pins 0, 1 and 6 of PORT C to output
    movlw 0xBC
    movwf TRISC, ACCESS
    ; set the green LED and clear the others
    clrf PORTC, ACCESS
    bsf PORTC, GREEN_LED_PIN, ACCESS

    ; set the function pointer to the USB vendor request handler
    movlw low(VendorRequestHandler)
    movwf USB_HANDLER_FUNCTIONL, ACCESS
    movlw high(VendorRequestHandler)
    movwf USB_HANDLER_FUNCTIONH, ACCESS
    movlw upper(VendorRequestHandler)
    movwf USB_HANDLER_FUNCTIONU, ACCESS
    bsf USB_CONTROL, USB_FIRMWARE_HANDLER_FLAG, ACCESS

    ; relocate the EP 0 buffer
    movlw low(EP0_BUFFER_START)
    movwf USB_CONTROL + 1, ACCESS
    movlw high(EP0_BUFFER_START)
    movwf USB_CONTROL + 2, ACCESS

    ; setup the USB module
    call InitializeUSB

    ; ----- firmware endpoint setup -----
    banksel BD2STAT
    movlw low(EP1_BUFFER_START)
    movwf BD2ADRL, BANKED
    movlw high(EP1_BUFFER_START)
    movwf BD2ADRH, BANKED
    movlw EP1_BUFFER_SIZE
    movwf BD2CNT, BANKED
    ; arm the endpoint
    movlw 0x88
    movwf BD2STAT, BANKED
    ; EP1 shall have OUT capability only
    movlw 0x1C
    movwf UEP1, ACCESS
    ; ----- firmware endpoint setup -----

    ; handle all USB requests until the host has finally configured the
    ; device
wait_for_USB_configuration_loop
    call HandleUSB
    ; get the status
    movf USB_CONTROL + 1, W, ACCESS
    andlw USB_STATUS_MASK
    xorlw USB_STATUS_CONFIGURED
    bnz wait_for_USB_configuration_loop

mainloop
    call HandleUSB
    goto mainloop

; ----------------------------------------------------------------------
; Handle a USB vendor request.
; Input:
;   FSR0 points to the beginning of the payload of the SETUP token if
;   the request belongs to endpoint 0. Otherwise it is not set.
; ----------------------------------------------------------------------
VendorRequestHandler
    ; find out which endpoint is wanted
    movf USTAT, W, ACCESS
    andlw 0x78
    bz vrh_handle_ep0
    ; this is not endpoint 0 -> test for endpoint 1
    xorlw (0x01 << 3)
    bnz vrh_unknown_endpoint
        ; this is endpoint 1
        ; make FSR0 point to the values we got
        lfsr 0, EP1_BUFFER_START
        ; ... at this point we could do something with the values ...
        ; arm the endpoint again
        banksel BD2STAT
        movlw EP1_BUFFER_SIZE
        movwf BD2CNT, BANKED
        movf BD2STAT, W, BANKED
        andlw 0x40                  ; remove everything but DATAx bit
        xorlw 0x40                  ; toggle DATAx
        iorlw 0x88                  ; UOWN & DTSEN
        movwf BD2STAT, BANKED
        return
vrh_handle_ep0
    ; the activity belongs to endpoint 0 -> FSR0 points to the payload
    ; of the SETUP token
    ; load bRequest into WREG
    movf PREINC0, W
    ; at this point we should dispatch the values of WREG (bRequest) but as
    ; there are no commands implemented, we go to the unknown command section
    ; immediately
vrh_unknown_command
vrh_unknown_endpoint
    ; issue an error
    bsf USB_CONTROL + 1, USB_ERROR_FLAG, ACCESS
    return

    end

; End of $Id: interrupt_transfer_firmware.asm 20 2010-05-04 22:33:20Z manuel $
