; $Id: project_template.asm 20 2010-05-04 22:33:20Z manuel $

; ==========================================================================
;
; (c) Manuel Freiberger
; 2008-2011
; manuel.freiberger@gmx.at
;
; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program.  If not, see <http://www.gnu.org/licenses/>.
;
; ==========================================================================
; Firmware template for student projects using the USB bootloader
;
; This is a template for projects using the USB bootloader. All projects
; must conform to the following specification:
; * idVendor and idProduct is a fixed pair. The values that have to be used
;   are idVendor = 0x1C40 and idProduct = 0x07EC.
; * The device must handle a USB vendor request called 'GET VERSION' with
;   bRequest = 0x00. If such a request is sent by the host, the device
;   will return a block of 8 byte. The meaning of these bytes is the
;   following (in the order as they are sent):
;   byte  0-3: Reserved. These bytes have to be set to 0.
;   byte  4-5: Project id. This is a unique number for the project. You'll
;              get this number from your advisor. The id is sent LSB first.
;   byte  6-7: The version of the project. The version has the form
;              'major.minor', where both major and minor have one byte.
;              minor is sent first, followed by major.
; * The device must also handle a USB vendor request called 'RESET' with
;   bRequest = 0xFA. The device does not acknowledge this USB transfer. It
;   will shutdown all activities, drive external components to a safe state
;   and reset itself.
; ==========================================================================

; Processor specific stuff
list p=18f2550
include <p18f2550.inc>

include "bootloader.inc"

; This is the unique 16-bit project id. You will get this number from your
; advisor.
#define PROJECT_ID      0xF00?
; The firmware version in the form major.minor (v1.0). Version 2.10 would be
; encoded as 0x020A, for example.
#define PROJECT_VERSION 0x0100

; ---------------------------------------------------------------------------
; USB vendor requests
; ---------------------------------------------------------------------------
; 0x00: GET VERSION
; Input:
;   None
; Output: 8 bytes
;   byte  0-3: Reserved. These bytes have to be set to 0.
;   byte  4-5: Project id. This is a unique number for the project. You'll
;              get this number from your advisor. The id is sent LSB first.
;   byte  6-7: The version of the device' firmware. The version has the
;              form 'major.minor', where both major and minor have one
;              byte. minor is sent first, followed by major.
VR_GET_VERSION EQU 0x00

; 0x01: SET LED
; Input:
;   wValue: 0         ... clear the LED
;           otherwise ... set the LED
; Output:
;   None
; Note:
;   This is an example command. You can implement your own commands in the
;   same fashion. The example assumes a LED connected to pin PORTC<1>
VR_SET_LED    EQU 0x01

; 0xFA: RESET
; Input:
;   None
; Output:
;   None. The transfer will not be acknowledged either because the device
;   will reset itself as fast as possible.
VR_RESET      EQU 0xFA


; ---------------------------------------------------------------------------
; Interrupt and reset vectors.
; ---------------------------------------------------------------------------
    org 0x0800
    goto FirmwareMain           ; offset 0x00
    goto MyGetUSBDescriptor     ;        0x04
    goto $                      ;        0x08
    nop                         ;        0x0C
    nop                         ;        0x0E
    nop                         ;        0x10
    nop                         ;        0x12
    nop                         ;        0x14
    nop                         ;        0x16
    goto $                      ;        0x18

; ---------------------------------------------------------------------------
; Custom function to fill the USB buffer with a USB descriptor.
;
; This function fills the USB buffer with a USB descriptor. The type of
; descriptor that is desired is stored in (USB_CONTROL + 4). The two most
; significant bits determine the descriptor type and the 6 least significant
; bit specify the descriptor index (e.g. 10000010 would stand for the
; 2nd string descriptor).
; The start offset (i.e. the index of the first byte which has to be copied)
; is stored in (USB_CONTROL + 3). The number of bytes that have to be copied
; can be found in (USB_CONTROL + 5).
; The function copies (USB_CONTROL + 5) bytes from the descriptor specified
; by (USB_CONTROL + 4) to INDF1 starting with the byte given by
; (USB_CONTROL + 3).
; When returning to the calling code, it is guaranteed, that the value in
; FSR1 and (USB_CONTROL + 3) have been increased by the value in
; (USB_CONTROL + 5), i.e. the pointer has advanced by the number of bytes
; copied and the offset has increased by the same amount. Finally, the
; value in (USB_CONTROL + 5) is guaranteed to be zero.
;
; Input:
;   USB_CONTROL + 3 ... Offset of the first byte of the descriptor to fetch.
;   USB_CONTROL + 4 ... Type of descriptor together with its index.
;     bits 7:6 - 00 ... device descriptor
;                01 ... configuration descriptor (index in bits 5:0)
;                10 ... string descriptor (index in bits 5:0)
;   USB_CONTROL + 5 ... Number of bytes to copy into the buffer. NOTE: This
;                       value can be zero and no byte must be copied in this
;                       case.
;   FSR1            ... Pointer to the output buffer where the descriptor has
;                       to be written to.
;
; Required output:
;   - The USB descriptor has been written to INDF1 (via FSR1).
;   - FSR1 and the register (USB_CONTROL + 3) have both been increased by the
;     number of bytes in register (USB_CONTROL + 5).
;   - Register (USB_CONTROL + 5) has been set to zero.
;
; Modified:
;   w, FSR1, TBLPTRx, TABLAT, USB_CONTROL + 3, USB_CONTROL + 5
; ---------------------------------------------------------------------------
MyGetUSBDescriptor
    movlw upper(MyGetUSBDescriptor)
    movwf TBLPTRU, ACCESS
    movlw high(MyGetUSBDescriptor)
    movwf TBLPTRH, ACCESS
    ; extract the descriptor type
    movf USB_CONTROL + 4, W, ACCESS
    andlw 0xC0
    bnz mgud_check_configuration
    ; the device descriptor is wanted
    movlw low(DEVICE_DESCRIPTOR)
    bra mgud_add_index_and_fetch_byte
mgud_check_configuration
    xorlw 0x40
    bnz mgud_check_string
    ; a configuration descriptor is wanted; get the index
    movf USB_CONTROL + 4, W, ACCESS
    andlw 0x3F
    ; we only have one configuration descriptor; if the index is not zero
    ; then we issue an error
    bnz mgud_issue_USB_error
    movlw low(CONFIGURATION1_DESCRIPTOR)
    bra mgud_add_index_and_fetch_byte
mgud_check_string
    ; a string descriptor is wanted; get the index
    movf USB_CONTROL + 4, W, ACCESS
    andlw 0x3F
    bnz mgud_check_string_1
    ; string 0 is wanted
    movlw low(STRING0_DESCRIPTOR)
    bra mgud_add_index_and_fetch_byte
mgud_check_string_1
    xorlw 0x01
    bnz mgud_check_string_2
    ; string 1 is wanted
    movlw low(STRING1_DESCRIPTOR)
    bra mgud_add_index_and_fetch_byte
mgud_check_string_2
    xorlw (0x01 ^ 0x02)
    bnz mgud_issue_USB_error
    ; string 2 is wanted
    movlw low(STRING2_DESCRIPTOR)
    bra mgud_add_index_and_fetch_byte
mgud_issue_USB_error
    bsf USB_CONTROL + 1, USB_ERROR_FLAG, ACCESS
    return
mgud_add_index_and_fetch_byte
    ; add index
    addwf USB_CONTROL + 3, W, ACCESS
    movwf TBLPTRL, ACCESS
    ; increase the index in (USB_CONTROL + 3) by the number of bytes we
    ; copy into the buffer which is stored in (USB_CONTROL + 5)
    movf USB_CONTROL + 5, W, ACCESS
    addwf USB_CONTROL + 3, F, ACCESS
    ; fetch the bytes in a loop
    incf USB_CONTROL + 5, F, ACCESS
mgud_loop
    dcfsnz USB_CONTROL + 5, F, ACCESS
    bra mgud_loop_done
    tblrd*+
    movf TABLAT, W, ACCESS
    movwf POSTINC1
    bra mgud_loop
mgud_loop_done
    return

; there has to be exactly one device descriptor for every device
; NOTE: a limitation of this bootloader is that only one configuration is
; supported which means that bNumConfigurations has to be 1
DEVICE_DESCRIPTOR
    db 0x12, 0x01           ; bLength, bDescriptorType
    db 0x00, 0x02           ; bcdUSB (2.0)
    db 0x00, 0x00           ; bDeviceClass (use interface class),bDeviceSubClass
       ; bDeviceProtocol (use interface protocol),
       ; bMaxPacketSize0 (low-speed: 8, full: 8/16/32/64)
    db 0x00, D'64'
    db 0x40, 0x1C           ; idVendor (0x1C40)
    db 0xEC, 0x07           ; idProduct (0x07EC)
    db 0x01, 0x00           ; bcdDevice (0.1)
    db 0x01, 0x02           ; iManufacturer, iProduct
    db 0x00, 0x01           ; iSerialnumber, bNumConfigurations
DEVICE_DESCRIPTOR_END

; there can be as many configuration descriptors per device as needed as long
; as the number of endpoints does not exceed 16; a configuration can have
; interfaces and endpoints (note that endpoint 0 does not need a descriptor
; as it is standardized)
; NOTE: a limitation of this bootloader is that the configuration descriptor
; must fit in 256 byte such that the high byte of wTotalLength is 0
CONFIGURATION1_DESCRIPTOR
    db 0x09, 0x02           ; bLength, bDescriptorType
                            ; wTotalLength
    db low(CONFIGURATION1_DESCRIPTOR_END - CONFIGURATION1_DESCRIPTOR), 0x00
    db 0x01, 0x01           ; bNumInterfaces, bConfigurationValue
    db 0x00, 0x80           ; iConfiguration, bmAttributes (bus-powered)
    ; after bMaxPower the CONFIG1_INTERFACE1_DESCRIPTOR starts
    db D'50', 0x09          ; bMaxPower (in units of 2mA), bLength
    db 0x04, 0x00           ; bDescriptorType, bInterfaceNumber
    db 0x00, 0x00           ; bAlternateSetting, bNumEndpoints without EP0
    db 0xFF, 0x00           ; bInterfaceClass (vendor), bInterfaceSubClass
    db 0xFF, 0x00           ; bInterfaceProtokoll (vendor), iInterface
CONFIG1_INTERFACE1_DESCRIPTOR_END
; endpoint 0 needs no descriptor; the others do and look like
;CONFIG1_ENDPOINT1_DESCRIPTOR
;    dt 0x07                 ; bLength
;    dt 0x05                 ; bDescriptorType
;    dt 0x01                 ; bEndpointAddress
;    dt 0x__                 ; bmAttributes
;    dt 0x__, 0x__           ; wMaxPacketSize
;    dt 0x__                 ; bInterval
;CONFIG1_ENDPOINT1_DESCRIPTOR_END
;CONFIG1_ENDPOINT2_DESCRIPTOR
;    dt ...etc...
;CONFIG1_ENDPOINT2_DESCRIPTOR_END
CONFIGURATION1_DESCRIPTOR_END

STRING0_DESCRIPTOR
    db 0x04, 0x03           ; bLength, bDescriptorType
    db 0x09, 0x04           ; wLANGID[0] (English - United States)
STRING0_DESCRIPTOR_END

; This is the manufacturer string (iManufacturer). Set it to whatever you like.
STRING1_DESCRIPTOR
                            ; bLength, bDescriptorType
    db low(STRING1_DESCRIPTOR_END - STRING1_DESCRIPTOR), 0x03
    db 'S', 0x00
    db 't', 0x00
    db 'u', 0x00
    db 'd', 0x00
    db 'e', 0x00
    db 'n', 0x00
    db 't', 0x00
STRING1_DESCRIPTOR_END

; This is the product string (iProduct). Set it to whatever you like.
STRING2_DESCRIPTOR
                            ; bLength, bDescriptorType
    db low(STRING2_DESCRIPTOR_END - STRING2_DESCRIPTOR), 0x03
    db 'T', 0x00
    db 'e', 0x00
    db 'm', 0x00
    db 'p', 0x00
    db 'l', 0x00
    db 'a', 0x00
    db 't', 0x00
    db 'e', 0x00
STRING2_DESCRIPTOR_END

    ; make sure that we did not cross a page boundary with the descriptor table
    if (high($) != high(MyGetUSBDescriptor))
        error "Descriptor definition crosses page boundary!"
    endif

; ---------------------------------------------------------------------------
; This is the beginning of the firmware.
; ---------------------------------------------------------------------------
FirmwareMain
    ; set pin 1 of PORT C to output
    bcf TRISC, TRISC1, ACCESS
    ; clear the LED
    bcf PORTC, 1, ACCESS

    ; set the function pointer to the USB vendor request handler
    movlw low(VendorRequestHandler)
    movwf USB_HANDLER_FUNCTIONL, ACCESS
    movlw high(VendorRequestHandler)
    movwf USB_HANDLER_FUNCTIONH, ACCESS
    movlw upper(VendorRequestHandler)
    movwf USB_HANDLER_FUNCTIONU, ACCESS
    ; enable our function handler
    bsf USB_CONTROL, USB_FIRMWARE_HANDLER_FLAG, ACCESS

    ; Relocate the USB buffer if needed. The following commands will place
    ; the USB buffer at 0x500. This has to be done prior to calling
    ; InitializeUSB. The buffer size is two times the package size of
    ; enpoint 0.
    movlw low(0x500)
    movwf USB_CONTROL + 1, ACCESS
    movlw high(0x500)
    movwf USB_CONTROL + 2, ACCESS

    ; setup the USB module
    call InitializeUSB
    ; handle all USB requests until the host has finally configured the
    ; device
wait_for_USB_configuration_loop
    call HandleUSB
    ; get the status
    movf USB_CONTROL + 1, W, ACCESS
    andlw USB_STATUS_MASK
    xorlw USB_STATUS_CONFIGURED
    bnz wait_for_USB_configuration_loop

mainloop
    ; handle USB events periodically
    call HandleUSB
    goto mainloop

; ---------------------------------------------------------------------------
; Handle a USB vendor request.
;
; Input:
;   FSR0 points to the beginning of the payload of the SETUP token.
; ---------------------------------------------------------------------------
VendorRequestHandler
    ; get bRequest (2nd byte in payload)
    movf PREINC0, W
    xorlw VR_GET_VERSION
    bnz vrh_check_set_led
        ; 'GET VERSION' was sent. Handling this command is mandatory for
        ; all projects.
        ; make FSR0 point to IN buffer
        movff BD1ADRL, FSR0L
        movff BD1ADRH, FSR0H
        ; bytes 0-3 are reserved and set to zero
        clrf POSTINC0
        clrf POSTINC0
        clrf POSTINC0
        clrf POSTINC0
        ; bytes 4-5 are the project id (lsb first)
        movlw low(PROJECT_ID)
        movwf POSTINC0
        movlw high(PROJECT_ID)
        movwf POSTINC0
        ; bytes 6-7 are the project version (minor is sent first)
        movlw low(PROJECT_VERSION)
        movwf POSTINC0
        movlw high(PROJECT_VERSION)
        movwf POSTINC0
        ; send 8 bytes
        banksel BD1CNT
        movlw 0x08
        movwf BD1CNT, BANKED
        movlw USB_TRANCEIVE_DATA1
        movwf BD1STAT, BANKED
        return
vrh_check_set_led
    xorlw (VR_GET_VERSION ^ VR_SET_LED)
    bnz vrh_check_reset
        ; 'SET LED' was sent
        ; get the desired state from wValue (3rd byte in payload)
        movf PREINC0, W
        ; if wValue == 0, clear the LED
        btfsc STATUS, Z, ACCESS
        bcf PORTC, 1, ACCESS
        ; if wValue != 0, set the LED
        btfss STATUS, Z, ACCESS
        bsf PORTC, 1, ACCESS
        ; acknowledge USB transfer by another transfer of 0 bytes (STATUS stage)
        banksel BD1CNT
        clrf BD1CNT, BANKED
        movlw USB_TRANCEIVE_DATA1
        movwf BD1STAT, BANKED
        return
vrh_check_reset
    xorlw (VR_SET_LED ^ VR_RESET)
    btfss STATUS, Z, ACCESS
    bra vrh_unknown_command
        ; 'RESET' was sent
        ; stop EP0 and the USB module
        clrf UEP0, ACCESS
        clrf UCON, ACCESS
        clrf UCFG, ACCESS
        ; LED off
        clrf PORTC, ACCESS
        ; reset the device
        reset
vrh_unknown_command
    ; issue an error
    bsf USB_CONTROL + 1, USB_ERROR_FLAG, ACCESS
    return

    end

; End of $Id: project_template.asm 20 2010-05-04 22:33:20Z manuel $
