// (c) Manuel Freiberger
// 2008-2011
// manuel.freiberger@gmx.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

// ==========================================================================
// Template for student projects
//
// This program can serve as a template for student projects.
// ==========================================================================

// $Id: project_template_host.cpp 22 2010-05-04 23:01:10Z manuel $

#include "usbdevice.hpp"

#include <iostream>

// Constants for the project. These values have to match the values specified
// in the firmware.
const unsigned PROJECT_ID      = 0xF00?;  // TODO: enter your number here
const unsigned PROJECT_VERSION = 0x0100;  // version 1.0

// example request for setting a LED
const int VENDOR_REQUEST_SET_LED = 0x01;

//! This class handles the USB communication with your device.
class ProjectUSBDevice : public USBDevice
{
  public:
    //! Constructor. Initializes internal state of the object but does not
    //! communicate over the USB, yet. Call \p init prior to any other
    //! USB communication. The timeout of USB transfers is set to \p timeout.
    ProjectUSBDevice(int timeout);

    //! The destructor will release the USB interface of the device.
    virtual ~ProjectUSBDevice();

    //! Reset the device.
    virtual void reset();

    //! Toggles the LED at PORT C pin 1. If successful, the method returns
    //! \p true.
    void toggleLED();

  private:
    //! State of the LED.
    bool m_led_state;
};

ProjectUSBDevice::ProjectUSBDevice(int timeout)
  : USBDevice(PROJECT_ID, PROJECT_VERSION, timeout),
    m_led_state(false)
{
}

ProjectUSBDevice::~ProjectUSBDevice()
{
}

void ProjectUSBDevice::reset()
{
  m_led_state = 0;
  USBDevice::reset();
}

void ProjectUSBDevice::toggleLED()
{
  int bytes_sent = libusb_control_transfer(
                     m_usb_device_handle,
                     VENDOR_REQUEST_TYPE_OUT, VENDOR_REQUEST_SET_LED,
                     !m_led_state, 0, 0, 0, m_timeout);
  if (bytes_sent != 0)
  {
    m_error_string = "Error while toggeling LED.";
    return;
  }
  // store the new state
  m_led_state = !m_led_state;
}

void printMenu();

int main(int argc, char* argv[])
{
  ProjectUSBDevice device(100);
  device.init();
  if (device.hasError())
  {
    std::cout << device.getErrorString() << std::endl;
    return -1;
  }

  std::cout << "USB device identification:" << std::endl;
  std::string manufacturer = device.getStringDescriptor(1);
  std::string product = device.getStringDescriptor(2);
  if (device.hasError())
  {
    std::cout << device.getErrorString() << std::endl;
    return -1;
  }
  std::cout << "  Manufacturer: " << manufacturer << std::endl;
  std::cout << "  Product: " << product << std::endl;
  std::cout << std::endl;

  printMenu();
  while (1)
  {
    char ch = std::cin.get();
    switch (ch)
    {
      case 'r':
        device.reset();
        if (device.hasError())
        {
          std::cout << device.getErrorString() << std::endl;
          return -1;
        }
        break;

      case 'i':
      {
        unsigned id = device.getId();
        if (device.hasError())
        {
          std::cout << device.getErrorString() << std::endl;
          return -1;
        }
        std::cout << "Id: " << std::hex << id << std::dec << std::endl;
      } break;

      case 'v':
      {
        unsigned version = device.getVersion();
        if (device.hasError())
        {
          std::cout << device.getErrorString() << std::endl;
          return -1;
        }
        std::cout << "Version: " << ((version >> 8) & 0xFF)
                  << '.' << (version & 0xFF) << std::endl;
      } break;

      case 't':
        device.toggleLED();
        if (device.hasError())
        {
          std::cout << device.getErrorString() << std::endl;
          return -1;
        }
        break;

      case 'q':
        return 0;

      default:
        break;
    }
  }
  return 0;
}

void printMenu()
{
  std::cout << "I   Get id of USB device" << std::endl;
  std::cout << "V   Get version of USB device" << std::endl;
  std::cout << "T   Toggle LED" << std::endl;
  std::cout << "R   Reset USB device" << std::endl;
  std::cout << "Q   Quit" << std::endl;
}

// End of $Id: project_template_host.cpp 22 2010-05-04 23:01:10Z manuel $
