// (c) Manuel Freiberger
// 2008-2011
// manuel.freiberger@gmx.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

// ==========================================================================
// Defines a class for USB communication with a device based on the
// USB-Bootloader
// ==========================================================================

// $Id: usbdevice.cpp 22 2010-05-04 23:01:10Z manuel $

#include "usbdevice.hpp"

USBDevice::USBDevice(unsigned project_id, unsigned project_version,
                     int timeout)
  : m_error_string(""),
    m_interface_claimed(false),
    m_project_id(project_id),
    m_project_version(project_version),
    m_timeout(timeout),
    m_usb_context(0),
    m_usb_device_handle(0)
{
}

USBDevice::~USBDevice()
{
  close();
  // close the lib
  if (m_usb_context)
    libusb_exit(m_usb_context);
}

void USBDevice::close()
{
  // release the interface if claimed
  if (m_interface_claimed && m_usb_device_handle)
  {
    libusb_release_interface(m_usb_device_handle, 0);
    m_interface_claimed = false;
  }
  // close the device
  if (m_usb_device_handle)
  {
    libusb_close(m_usb_device_handle);
    m_usb_device_handle = 0;
  }
}

const std::string& USBDevice::getErrorString() const
{
  return m_error_string;
}

unsigned USBDevice::getId() const
{
  return m_project_id;
}

unsigned USBDevice::getVersion() const
{
  return m_project_version;
}

std::string USBDevice::getStringDescriptor(unsigned index)
{
  unsigned char str_desc[256];
  int bytes_read = libusb_get_string_descriptor_ascii(
                     m_usb_device_handle, index, str_desc, 255);
  if (bytes_read >= 0)
  {
    return std::string((char*)str_desc, bytes_read);
  }
  else
  {
    m_error_string = "Error while reading string descriptor.";
    return "";
  }
}

bool USBDevice::hasError() const
{
  return !m_error_string.empty();
}

void USBDevice::init(bool match_project_version)
{
  // initialize the USB library
  if (m_usb_context == 0)
  {
    if (libusb_init(&m_usb_context) != 0)
    {
      m_error_string = "Could not initialize libusb.";
      return;
    }
    libusb_set_debug(m_usb_context, 3);
  }
  // get all devices
  libusb_device** list;
  ssize_t num_devices = libusb_get_device_list(m_usb_context, &list);
  if (num_devices < 0)
  {
    m_error_string = "Could not get a list of USB devices.";
    return;
  }

  // now look for our device
  for (ssize_t counter = 0; counter < num_devices; ++counter)
  {
    // get a device descriptor for the current device
    libusb_device_descriptor descriptor;
    if (libusb_get_device_descriptor(list[counter], &descriptor) != 0)
      continue;

    // check vendor and product id
    if ((descriptor.idVendor != 0x1C40) || (descriptor.idProduct != 0x07EC))
      continue;

    // at this point we have a device that is from a project; we have to
    // compare the project id and the version now;
    // but first try to open the device...
    libusb_device_handle* test_device = 0;
    if (libusb_open(list[counter], &test_device) != 0 || !test_device)
      continue;
    // try to claim interface 0
    if (libusb_claim_interface(test_device, 0) != 0)
    {
      // it did not work -> close the device and go on with the next one
      libusb_close(test_device);
      continue;
    }
    // get the version of the device
    std::vector<unsigned char> result;
    if (!getIdVersion(test_device, result) || result.size() != 8)
    {
      // could not get the version -> release the interface and the device
      libusb_release_interface(test_device, 0);
      libusb_close(test_device);
      continue;
    }
    bool match = true;
    std::vector<unsigned char>::const_iterator iter = result.begin();
    // bytes 0-3 have to be zero
    for (unsigned counter = 0; counter < 4; ++counter)
      match = match && (*iter++ == 0);
    // compare project id (lsb first)
    match = match && (*iter++ == (m_project_id & 0xFF));
    match = match && (*iter++ == ((m_project_id >> 8) & 0xFF));
    if (match_project_version)
    {
      // compare project version (minor first)
      match = match && (*iter++ == (m_project_version & 0xFF));
      match = match && (*iter++ == ((m_project_version >> 8) & 0xFF));
    }
    else
    {
      // do not match the project version but store it
      m_project_version = unsigned(*iter++) + (unsigned(*iter++) << 8);
    }
    // test if all comparisons worked
    if (!match)
    {
      // matching failed -> release the interface and the device
      libusb_release_interface(test_device, 0);
      libusb_close(test_device);
      continue;
    }
    // at this point we have a valid open device
    m_usb_device_handle = test_device;
    m_interface_claimed = true;
    libusb_free_device_list(list, 1);
    return;
  }
  // no matching device
  libusb_free_device_list(list, 1);
  m_error_string = "No matching USB device found. Possible reasons are:\n"
                   "  -The device is not connected or is not powered.\n"
                   "  -Lacking permissions for accessing USB.\n"
                   "  -Project id and/or version do not match.";
}

void USBDevice::reset()
{
  // reset the device; this command will timeout because the device resets
  // ifself and cannot respond to USB commands any longer!
  libusb_set_debug(m_usb_context, 0);
  libusb_control_transfer(m_usb_device_handle,
                          VENDOR_REQUEST_TYPE_OUT, VENDOR_REQUEST_RESET,
                          0, 0, 0, 0, m_timeout);
  // try a simple reset
  if (libusb_reset_device(m_usb_device_handle))
    return;
  // if the simple reset did not work, close the device and re-open it
  close();
  libusb_set_debug(m_usb_context, 3);
  init();
}

void USBDevice::unsetError()
{
  m_error_string = "";
}

bool USBDevice::getIdVersion(libusb_device_handle* device,
                             std::vector<unsigned char>& result)
{
  result.resize(8);
  int bytes_read = libusb_control_transfer(
                     device,
                     VENDOR_REQUEST_TYPE_IN, VENDOR_REQUEST_GET_VERSION,
                     0, 0, &result[0], 8, m_timeout);
  return (bytes_read == 8);
}

// End of $Id: usbdevice.cpp 22 2010-05-04 23:01:10Z manuel $
