// (c) Manuel Freiberger
// 2008-2011
// manuel.freiberger@gmx.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

// ==========================================================================
// Defines a class for USB communication with a device based on the
// USB-Bootloader
// ==========================================================================

// $Id: usbdevice.hpp 22 2010-05-04 23:01:10Z manuel $

#ifndef USBDEVICE_HPP
#define USBDEVICE_HPP

#include <libusb-1.0/libusb.h>
#include <string>
#include <vector>

// Constants for USB vendor requests. The requests 0x00 and 0xFA are mandatory.
const int VENDOR_REQUEST_GET_VERSION = 0x00;
const int VENDOR_REQUEST_RESET       = 0xFA;

// abbreviations for USB request directions
const int VENDOR_REQUEST_TYPE_OUT = 0x40;
const int VENDOR_REQUEST_TYPE_IN  = 0xC0;

//! This class handles the USB communication with a USB bootloader device.
class USBDevice
{
  public:
    //! Constructor. Initializes internal state of the object but does not
    //! communicate over the USB, yet. Call \p init() prior to any other
    //! USB communication. The constructor takes the project \p id and
    //! project \p version which are needed to identify the correct USB device.
    //! The timeout of USB transfers is set to \p timeout.
    USBDevice(unsigned project_id, unsigned project_version, int timeout);

    //! The destructor will release the USB interface of the device.
    virtual ~USBDevice();

    //! Close the USB device. No further USB communication is allowed after
    //! calling this method.
    virtual void close();

    //! Return the error string. If there was no error, this method will
    //! return an empty string.
    virtual const std::string& getErrorString() const;

    //! This method returns the project id which was supplied to the
    //! constructor.
    virtual unsigned getId() const;

    //! This method retrieves the project version. If project matching when
    //! calling \p init() is turned off, this method will return the project
    //! version as returned by the USB device's firmware.
    virtual unsigned getVersion() const;

    //! Get the \p index-th string descriptor from the USB device.
    virtual std::string getStringDescriptor(unsigned index);

    //! If a USB communication fails, subsequent calls to this method will
    //! return \p true. \p getErrorString() will return a description of
    //! the error.
    virtual bool hasError() const;

    //! Initialize USB communication. This method initializes the USB bus and
    //! looks for a device with matching idVendor, idProduct, project ID
    //! and project version. If \p match_project_version is set to \p false,
    //! no match against the version is performed. In this case, the method
    //! will accept the first device encountered which has a matching
    //! vendor id, product id and project id. The project version can be
    //! queried using \p getVersion().
    virtual void init(bool match_project_version = true);

    //! Reset the device.
    virtual void reset();

    //! Clear the error state.
    virtual void unsetError();

  protected:
    //! The error string.
    std::string m_error_string;

    //! Timeout for USB transfers.
    int m_timeout;

    //! USB context.
    libusb_context* m_usb_context;

    //! USB handle to our device.
    libusb_device_handle* m_usb_device_handle;

  private:
    //! Flag to store if the interface was claimed.
    bool m_interface_claimed;

    //! The project id of this USB device.
    unsigned m_project_id;

    //! The project version of this USB device.
    unsigned m_project_version;

    //! Hidden default constructor.
    USBDevice();

    //! Method to get the id and version of a certain \p device. \p result
    //! is an array holding 8 bytes. The first 4 bytes will be zero, followed
    //! by the 16 bit project id (lsb first) and the 16 bit project version
    //! (lsb first).
    bool getIdVersion(libusb_device_handle* device,
                      std::vector<unsigned char>& result);
};

#endif

// End of $Id: usbdevice.hpp 22 2010-05-04 23:01:10Z manuel $
